#include "colors.inc"
#include "shapes.inc"
#include "lionfish.inc"
#include "lionfin.inc"
#include "eyespine.inc"
#include "eyeflap.inc"
#include "nosespn.inc"
#include "coral.inc"
#include "coralhd.inc"

camera {
  location  <1 0 -17>
  direction <0 0 1.5>
  up        <0 1 0>
  right     <1.33 0 0>
  look_at   <1 0 2>
}

object { light_source { <2 8 -6> color White } }

#declare Eyeball =
    union {
      quadric {
        Ellipsoid
        scale <.25 .175 .25>
        texture {
          wood
          turbulence .15
          scale <.2 .2 .2>
          color_map {
            [0.0 0.2 color White color White]
            [0.2 0.3 color Black color Black]
            [0.3 0.7 color ReddishBrown color ReddishBrown]
            [0.7 0.8 color Black color Black]
            [0.8 1.0 color White color White]
          }
        }
      }
      quadric {
        Ellipsoid 
        scale <.1 .025 .1>
        translate <0 .2 0>
        texture {color Black}
      }
    }

#declare Lionwing = // you can try using this if you want. 
                    //It wouldn't work for me. Use in place of 
                    //objects at lines 190 and 208.
object {
  height_field {gif "lionwing.gif" water_level 1}
  scale <8.5 .1 4.25>
  rotate <-90 0 0>
  texture {
    onion
    color_map {
      [0.0 0.25 color White color ReddishBrown]
      [0.25 0.5 color ReddishBrown color ReddishBrown]
      [0.50 0.75 color ReddishBrown color White]
      [0.75 1.0 color White color White]
    }
    scale <1 1 1>
    ambient 0.3
    diffuse 0.7
  }
}


#declare Eye_Spine_Flap =  
  composite {
    composite {
      Eyespine
      scale <.15 .125 .2>
    }
    composite {
      Eyeflap
      scale <.15 .125 .002>
      rotate <0 0 15>
      translate <0 .425 0>
    }
  }

object { //sandy bottom
  plane {<0 1 0> -4.5}
  texture {
    color Sand
    ambient 0.3
    diffuse 0.6
    ripples .35
    frequency 100
    turbulence 0.75
    scale <100 100 100>
  }
}

object { //the deep blue sea
  sphere {<0 0 0> 1000}
  texture {color SlateBlue}
}

fog {
  color SlateBlue
  200
} 

composite { //a coral branch
  Coral_Tree
  rotate <0 0 -10>
  scale <.6 .75 .6>
  translate <5 -3 10>
}

composite { //a coral head
  Coral_Head
  scale <1 1 1>
  translate <-9 -2 15>
}

composite { //the lionfish
  composite { //lionfish's body
    Lionfish_Body
    scale <.5 .45 .2>
  }
  composite { //caudal fin
    Lionfin
    scale <.5 .5 .01>
    translate <7 0 0>
  }
  composite { //rear dorsal fin
    Lionfin
    scale <.35 .25 .01>
    rotate <0 0 10>
    translate <4.5 1 0>
  }
  composite { //anal fin
    Lionfin
    scale <.35 .25 .01>
    rotate <0 0 -10>
    translate <4.5 -1 0>
  }
  composite { //left pectoral fin
    Lionfin
    scale <.35 .25 .01>
    rotate <0 0 -45>
    rotate <25 0 0>
    translate <0 -1.5 -.5>
  }
  composite { //right pectoral fin
    Lionfin
    scale <.35 .25 .01>
    rotate <0 0 -45>
    rotate <-25 0 0>
    translate <0 -1.5 .5>
  }
  object { //the lower jaw
    quadric {
      Ellipsoid
      scale <.6 .25 .25>
      rotate <0 0 -60>
      translate <-2.5 -.2 0>
    }
    texture {color White}
  }
  object { //left eye
    union {
      Eyeball
      scale <1.25 1.25 1.25>
      rotate <-75 0 0>
      translate <-1.5 .7 -.3>
    }
  }
  object { //right eye
    union {
      Eyeball
      scale <1.25 1.25 1.25>
      rotate <75 0 0>
      translate <-1.5 .7 .3>
    }
  }
  composite { //left eye spine
    Eye_Spine_Flap
    scale <.75 .75 .75>
    rotate <0 0 15>
    translate <-1.6 1.15 -.3>
  }
  composite { //right eye spine
    Eye_Spine_Flap
    scale <.75 .75 .75>
    rotate <0 0 15>
    translate <-1.6 1.15 .3>
  }
  object { //left pectoral "wing" fin 
    height_field {gif "lionwing.gif" water_level 1}
    scale <8.5 .025 4.25>
    rotate <-100 0 0>
    texture {
      onion 
      color_map {
        [0.0 0.25 color White color ReddishBrown]
        [0.25 0.5 color ReddishBrown color ReddishBrown]
        [0.5 0.75 color ReddishBrown color White]
        [0.75 1.0 color White color White]
      }
      ambient .3
      diffuse .7
    }
    rotate <-10 40 -15>
    translate <-1.25 -.5 .5>
  }
  object { //right pectoral "wing" fin
    height_field {gif "lionwing.gif" water_level 1}
    scale <8.5 .025 4.25>
    rotate <-60 0 0>
    texture {
      onion
      color_map {
        [0.0 0.25 color White color ReddishBrown]
        [0.25 0.5 color ReddishBrown color ReddishBrown]
        [0.5 0.75 color ReddishBrown color White]
        [0.75 1.0 color White color White]
      }
      ambient .3
      diffuse .7
    }
    rotate <10 -40 -15>
    translate <-.25 -.5 -.4>
  }
  object { //the dorsal spines
    height_field {gif "lionspin.gif" water_level 1}
    scale <8.5 .025 5>
    rotate <-90 0 0>
    texture {
      bozo
      color_map {
        [0.0 0.25 color White color ReddishBrown]
        [0.25 0.5 color ReddishBrown color ReddishBrown]
        [0.5 0.75 color ReddishBrown color White]
        [0.75 1.0 color White color White]
      }
      scale <.75 .25 .25>
      ambient .3
      diffuse .7
    }
    translate <-2.85 .25 0>
  }
  composite { the spine over the left eye
    Eye_Spine_Flap
    scale <.35 .35 .35>
    rotate <-140 0 -20>
    translate <-2.1 -.5 -.25>
  }
  composite { //the spine over the right eye
    Eye_Spine_Flap
    scale <.3 .3 .3>
    rotate <-140 0 0>
    translate <-1.9 -.6 -.25>
  }
  composite { //small spines by upper jaw
    Eye_Spine_Flap
    scale <.4 .4 .4>
    rotate <-140 0 -60>
    translate <-2.25 -.3 -.25>
  }
  composite {
    Eye_Spine_Flap
    scale <.35 .35 .35>
    rotate <140 0 -20>
    translate <-2.1 -.5 .25>
  }
  composite {
    Eye_Spine_Flap
    scale <.3 .3 .3>
    rotate <140 0 0>
    translate <-1.9 -.6 .25>
  }
  composite {
    Eye_Spine_Flap
    scale <.4 .4 .4>
    rotate <140 0 -60>
    translate <-2.25 -.3 .25>
  }
  composite { small spines at tip of nose
    Nose_Spine
    scale <.2 .125 .2>
    rotate <0 0 20>
    translate <-2.1 .55 0>
  }
  composite {
    Nose_Spine
    scale <.2 .125 .2>
    rotate <0 0 20>
    translate <-2.5 .445 0>
  }
  rotate <-30 0 0> 
  rotate <0 -20 0> 
  rotate <0 0 15>
  translate <-2 0 0>
}

